@@ -10,18 +10,24 @@ typedef struct {
     ffi_type **argv;
 } fiddle_closure;
 
-#if defined(MACOSX) || defined(__linux) || defined(__OpenBSD__)
-#define DONT_USE_FFI_CLOSURE_ALLOC
+#if defined(USE_FFI_CLOSURE_ALLOC)
+#elif defined(__OpenBSD__) || defined(__APPLE__) || defined(__linux__)
+# define USE_FFI_CLOSURE_ALLOC 0
+#elif defined(RUBY_LIBFFI_MODVERSION) && RUBY_LIBFFI_MODVERSION < 3000005 && \
+	(defined(__i386__) || defined(__x86_64__) || defined(_M_IX86) || defined(_M_AMD64))
+# define USE_FFI_CLOSURE_ALLOC 0
+#else
+# define USE_FFI_CLOSURE_ALLOC 1
 #endif
 
 static void
 dealloc(void * ptr)
 {
     fiddle_closure * cls = (fiddle_closure *)ptr;
-#ifndef DONT_USE_FFI_CLOSURE_ALLOC
+#if USE_FFI_CLOSURE_ALLOC
     ffi_closure_free(cls->pcl);
 #else
-    munmap(cls->pcl, sizeof(*cls->pcl));
+    munmap(cls->pcl, sizeof(cls->pcl));
 #endif
     if (cls->argv) xfree(cls->argv);
     xfree(cls);
@@ -170,7 +176,7 @@ allocate(VALUE klass)
     VALUE i = TypedData_Make_Struct(klass, fiddle_closure,
 	    &closure_data_type, closure);
 
-#ifndef DONT_USE_FFI_CLOSURE_ALLOC
+#if USE_FFI_CLOSURE_ALLOC
     closure->pcl = ffi_closure_alloc(sizeof(ffi_closure), &closure->code);
 #else
     closure->pcl = mmap(NULL, sizeof(ffi_closure), PROT_READ | PROT_WRITE,
@@ -222,13 +228,13 @@ initialize(int rbargc, VALUE argv[], VALUE self)
     if (FFI_OK != result)
 	rb_raise(rb_eRuntimeError, "error prepping CIF %d", result);
 
-#ifndef DONT_USE_FFI_CLOSURE_ALLOC
+#if USE_FFI_CLOSURE_ALLOC
     result = ffi_prep_closure_loc(pcl, cif, callback,
 		(void *)self, cl->code);
 #else
     result = ffi_prep_closure(pcl, cif, callback, (void *)self);
     cl->code = (void *)pcl;
-    i = mprotect(pcl, sizeof(*pcl), PROT_READ | PROT_EXEC);
+    i = mprotect(pcl, sizeof(pcl), PROT_READ | PROT_EXEC);
     if (i) {
 	rb_sys_fail("mprotect");
     }
@@ -274,11 +280,11 @@ Init_fiddle_closure()
      *       10
      *     end
      *   }.new(Fiddle::TYPE_INT, [])
-     *   => #<#<Class:0x0000000150d308>:0x0000000150d240>
+     *	    #=> #<#<Class:0x0000000150d308>:0x0000000150d240>
      *   func = Fiddle::Function.new(closure, [], Fiddle::TYPE_INT)
-     *   => #<Fiddle::Function:0x00000001516e58>
+     *	    #=> #<Fiddle::Function:0x00000001516e58>
      *   func.call
-     *   => 10
+     *	    #=> 10
      */
     cFiddleClosure = rb_define_class_under(mFiddle, "Closure", rb_cObject);
 
@@ -292,7 +298,7 @@ Init_fiddle_closure()
      * Construct a new Closure object.
      *
      * * +ret+ is the C type to be returned
-     * * +args+ are passed the callback
+     * * +args+ is an Array of arguments, passed to the callback function
      * * +abi+ is the abi of the closure
      *
      * If there is an error in preparing the ffi_cif or ffi_prep_closure,
