@@ -1,21 +1,24 @@
 require "rubygems"
 require "rubygems/test_case"
 require "rubygems/commands/help_command"
-require "rubygems/format"
+require "rubygems/package"
 require "rubygems/command_manager"
-require_relative "rubygems_plugin"
+require File.expand_path('../rubygems_plugin', __FILE__)
 
 class TestGemCommandsHelpCommand < Gem::TestCase
   def setup
     super
 
     @cmd = Gem::Commands::HelpCommand.new
+
+    load File.expand_path('../rubygems_plugin.rb', __FILE__) unless
+      Gem::Commands.const_defined? :InterruptCommand
   end
 
   def test_gem_help_bad
     util_gem 'bad' do |out, err|
       assert_equal('', out)
-      assert_match(/Unknown command bad. Try gem help commands\n/, err)
+      assert_match "Unknown command bad", err
     end
   end
 
@@ -33,7 +36,12 @@ class TestGemCommandsHelpCommand < Gem::TestCase
       mgr.command_names.each do |cmd|
         assert_match(/\s+#{cmd}\s+\S+/, out)
       end
-      assert_equal '', err
+
+      if defined?(OpenSSL::SSL) then
+        assert_empty err
+
+        refute_match 'No command found for ', out
+      end
     end
   end
 
